package ssw.mj.test;

import static ssw.mj.Errors.Message.*;

import org.junit.Test;

public class ParserErrorTest extends CompilerTestCase {
	
//  ---- Errors in Tab.java
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void tooManyLocals() {
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < 127; i++) {
			if (i > 0) {
				sb.append(",");
			}
			sb.append("i");
			sb.append(i);
		}
		String names = sb.toString();
		
		init("program Test {" + LF +
		     "  void main()" + LF +
		     "    int " + names + ";" + LF +
		     "    int error;" + LF +
		     "  {" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 4, 5, 3, TOO_MANY_LOCALS);
		parseAndVerify();
	}

	// this test case requires much time, so it is disabled by default
	@Test()
	public void tooManyGlobals() {
		if (!Configuration.RUN_SLOW_TESTS) {
			return;
		}
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < 32767; i++) {
			if (i > 0) {
				sb.append(",");
			}
			sb.append("i");
			sb.append(i);
		}
		String names = sb.toString();
		
		init("program Test" + LF +
		     "  int " + names + ";" + LF +
		     "  int error;" + LF +
		     "{" + LF +
		     "  void main() { }" + LF +
		     "}");
		expectError(EX >= 4, 4, 1, TOO_MANY_GLOBALS);
		
		parseAndVerify();
	}
	
	// this test case requires much time, so it is disabled by default
	@Test()
	public void tooManyFields() {
		if (!Configuration.RUN_SLOW_TESTS) {
			return;
		}
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < 32767; i++) {
			if (i > 0) {
				sb.append(",");
			}
			sb.append("i");
			sb.append(i);
		}
		String names = sb.toString();
		
		init("program Test" + LF +
		     "  class C {" + LF +
		     "    int " + names + ";" + LF +
		     "    int error;" + LF +
		     "  }" + LF +
		     "{" + LF +
		     "  void main() { }" + LF +
		     "}");
		expectError(EX >= 4, 5, 3, TOO_MANY_FIELDS);
		
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void doubleDeclVar() {
		init("program Test " + LF +
		     "  int x;" + LF +
		     "  int x;" + LF +
		     "{" + LF +
		     "  void main() { }" + LF +
		     "}");
		expectError(EX >= 4, 3, 8, DECL_NAME, "x");
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void doubleDeclMeth() {
		init("program Test " + LF +
		     "  int x;" + LF +
		     "{" + LF +
		     "  void x() { }" + LF +
		     "  void main() { }" + LF +
		     "}");
		expectError(EX >= 4, 4, 9, DECL_NAME, "x");
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void doubleDeclLocal() {
		init("program Test {" + LF +
		     "  void method(int x)" + LF +
		     "    int x;" + LF +
		     "  { }" + LF +
		     "  void main() { }" + LF +
		     "}");
		expectError(EX >= 4, 3, 10, DECL_NAME, "x");
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void undefNameType() {
		init("program Test" + LF +
		     "  type x;" + LF +
		     "{" + LF +
		     "  void main() { }" + LF +
		     "}");
		expectError(EX >= 4, 2, 8, NOT_FOUND, "type");
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void undefNameMeth() {
		init("program Test {" + LF +
		     "  void main() {" + LF +
		     "    method();" + LF + 
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 11, NOT_FOUND, "method");
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void undefNameVar() {
		init("program Test {" + LF +
		     "  void main() {" + LF +
		     "    var++;" + LF + 
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 8, NOT_FOUND, "var");
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noField() {
		init("program Test" + LF +
		     "  class C { }" + LF +
		     "{" + LF +
		     "  void main() C obj; {" + LF +
		     "    obj.field++;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 5, 14, NO_FIELD, "field");
		parseAndVerify();
	}
	
//   ---- Errors in Code.java
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noVarMethod() {
		init("program Test {" + LF +
		     "  int method() { return 0; }" + LF +
		     "  void main() int i; {" + LF +
		     "    method = i;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 4, 15, NO_VAR);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noVarIncMethod() {
		init("program Test {" + LF +
		     "  int method() { return 0; }" + LF +
		     "  void main() int i; {" + LF +
		     "    method++;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 4, 11, NO_VAR);
		parseAndVerify();
	}

	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noItem() {
		init("program Test {" + LF +
		     "  void main() int i; {" + LF +
		     "    Test = i;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 10, NO_ITEM);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noValueAssign() {
		init("program Test {" + LF +
		     "  char method() { return 'a'; }" + LF +
		     "  void main() char c; {" + LF +
		     "    c = method;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 4, 15, NO_VAL);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noValueCalc() {
		init("program Test {" + LF +
		     "  int method() { return 0; }" + LF +
		     "  void main() int i; {" + LF +
		     "    i = 5 * method;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 4, 19, NO_VAL);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noValueInc() {
		init("program Test {" + LF +
		     "  int method() { return 0; }" + LF +
		     "  void main() int i; {" + LF +
		     "    i += method;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 4, 16, NO_VAL);
		parseAndVerify();
	}
	
//	 ---- Errors in Parser.java

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongStart() {
		init("noprogram Test { }");
		expectError(EX >= 3, 1, 1, TOKEN_EXPECTED, "program");
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noProgName() {
		init("program { }");
		expectError(EX >= 3, 1, 9, TOKEN_EXPECTED, "identifier");
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongGlobalDecl () {
		init("program Test" + LF +
		     "  123;" + LF +
		     "{ void main() { } }");
		expectError(EX == 3, 2, 3, TOKEN_EXPECTED, "{");
		expectError(EX >= 4, 2, 3, INVALID_DECL);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noMain() {
		init("program Test {" + LF +
		     "  void main_() { }" + LF +
		     "}");
		expectError(EX >= 6, 3, 1, MAIN_NOT_FOUND);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void mainVar() {
		init("program Test" + LF +
		     "  int main;" + LF +
		     "{" + LF +
		     "}");
		expectError(EX >= 6, 4, 1, MAIN_NOT_FOUND);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongConstDecl() {
		init("program Test" + LF +
		     "  final int i = a;" + LF +
		     "{ void main() { } }");
		expectError(EX >= 3, 2, 17, CONST_DECL);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongConstTypeInt() {
		init("program Test" + LF +
		     "  final int i = 'a';" + LF +
		     "{ void main() { } }");
		expectError(EX >= 4, 2, 17, CONST_TYPE);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongConstTypeChar() {
		init("program Test" + LF +
		     "  final char ch = 32;" + LF +
		     "{ void main() { } }");
		expectError(EX >= 4, 2, 19, CONST_TYPE);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongConstType() {
		init("program Test" + LF +
			 "  class C { int i; }" + LF +
		     "  final C c = 32;" + LF +
		     "{ void main() { } }");
		expectError(EX >= 4, 3, 15, CONST_TYPE);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongMethDecl1() {
		init("program Test {" + LF +
		     "  void main() { }" + LF +
		     "  program wrong1() { " + LF +
		     "    while (1>2);" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX == 3, 3, 3, TOKEN_EXPECTED, "}");
		expectError(EX >= 4, 3, 3, METH_DECL);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongMethDecl2() {
		init("program Test {" + LF +
		     "  program wrong1() { " + LF +
		     "    while (1>2);" + LF +
		     "  }" + LF +
		     "  void main() { }" + LF +
		     "  program wrong2() {" + LF +
		     "    while (1>2);" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX == 3, 2, 3, TOKEN_EXPECTED, "}");
		expectError(EX >= 4, 2, 3, METH_DECL);
		expectError(EX >= 4, 6, 3, METH_DECL);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongMethDecl3() {
		init("program Test {" + LF +
		     "  program wrong1() { }" + LF +
		     "  void main() { }" + LF +
		     "  program wrong2() { }" + LF +
		     "}");
		expectError(EX == 3, 2, 3, TOKEN_EXPECTED, "}");
		expectError(EX >= 4, 2, 3, METH_DECL);
		expectError(EX >= 4, 4, 3, METH_DECL);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void mainNotVoid() {
		init("program Test {" + LF +
		     "  char main() { }" + LF +
		     "}");
		expectError(EX >= 6, 2, 15, MAIN_NOT_VOID);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void mainWithParams() {
		init("program Test {" + LF +
		     "  void main(int i) { }" + LF +
		     "}");
		expectError(EX >= 6, 2, 20, MAIN_WITH_PARAMS);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noType() {
		init("program Test" + LF +
		     "  int i;" + LF +
		     "  i s;" + LF +
		     "{ void main() { } }");
		expectError(EX >= 4, 3, 5, NO_TYPE);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void assignPlusNoIntOp() {
		init("program Test {" + LF +
		     "  void main() int i; char c; {" + LF +
		     "    c += i;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 11, NO_INT_OP);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void assignTimesNoIntOp() {
		init("program Test {" + LF +
		     "  void main() int i; char c; {" + LF +
		     "    i *= c;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 11, NO_INT_OP);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void incompTypes() {
		init("program Test {" + LF +
		     "  void main() int i; {  " + LF +
		     "    i = null;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 13, INCOMP_TYPES);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void incompTypesArr() {
		init("program Test {" + LF +
		     "  void main() int[] ia; char[] ca; {  " + LF +
		     "    ia = ca;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 12, INCOMP_TYPES);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void incompTypesClass() {
		init("program Test" + LF +
		     "  class C1 { }" + LF +
		     "  class C2 { }" + LF +
		     "{" + LF +
		     "  void main() C1 c1; C2 c2; {  " + LF +
		     "    c1 = c2;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 6, 12, INCOMP_TYPES);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noIntegerInc() {
		init("program Test {" + LF +
		     "  void main() char ch; {" + LF +
		     "    ch++;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 7, NO_INT);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noIntegerDec() {
		init("program Test {" + LF +
		     "  void main() int[] ia; {" + LF +
		     "    ia--;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 7, NO_INT);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongDesignFollow() {
		init("program Test {" + LF +
		     "  void main() int i; {" + LF +
		     "    i**;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 3, 3, 6, DESIGN_FOLLOW);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noLoop() {
		init("program Test {" + LF +
		     "  void main() {" + LF +
		     "    break;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 6, 3, 10, NO_LOOP);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void returnVoid() {
		init("program Test {" + LF +
		     "  void test() {" + LF +
		     "    return 5;" + LF +
		     "  }" + LF +
		     "  void main() {}" + LF +
		     "}");
		expectError(EX >= 6, 3, 12, RETURN_VOID);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongReturnType() {
		init("program Test {" + LF +
		     "  int test() {" + LF +
		     "    return null;" + LF +
		     "  }" + LF +
		     "  void main() {}" + LF +
		     "}");
		expectError(EX >= 6, 3, 16, RETURN_TYPE);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongReturnTypeArr() {
		init("program Test {" + LF +
		     "  int[] test() char[] ca; {" + LF +
		     "    return ca;" + LF +
		     "  }" + LF +
		     "  void main() {}" + LF +
		     "}");
		expectError(EX >= 6, 3, 14, RETURN_TYPE);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongReturnClass() {
		init("program Test" + LF +
		     "  class C1 { }" + LF +
		     "  class C2 { }" + LF +
		     "{" + LF +
		     "  C1 test() C2 c2; {" + LF +
		     "    return c2;" + LF +
		     "  }" + LF +
		     "  void main() {}" + LF +
		     "}");
		expectError(EX >= 6, 6, 14, RETURN_TYPE);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noReturnVal() {
		init("program Test {" + LF +
		     "  int test() {" + LF +
		     "    return;" + LF +
		     "  }" + LF +
		     "  void main() {}" + LF +
		     "}");
		expectError(EX >= 6, 3, 11, RETURN_NO_VAL);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongReadValue() {
		init("program Test {" + LF +
		     "  void main() int[] ia; { " + LF +
		     "    read(ia);" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 12, READ_VALUE);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongPrintValue() {
		init("program Test" + LF +
		     "  class C { }" + LF +
		     "{" + LF +
		     "  void main() C obj; { " + LF +
		     "    print(obj);" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 5, 14, PRINT_VALUE);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongStat() {
		init("program Test {" + LF +
		     "  void main() {  " + LF +
		     "    123;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX == 3, 3, 5, TOKEN_EXPECTED, "}");
		expectError(EX >= 4, 3, 5, INVALID_STAT);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noMeth() {
		init("program Test {" + LF +
		     "  void main() int i; {" + LF +
		     "    i(10);" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 6, 3, 7, NO_METH);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void paramType() {
		init("program Test {" + LF +
		     "  void method(int x) { }" + LF +
		     "  void main() {" + LF +
		     "    method('a');" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 6, 4, 15, PARAM_TYPE);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void paramTypeArr() {
		init("program Test {" + LF +
		     "  void method(int[] x) { }" + LF +
		     "  void main() {" + LF +
		     "    method(new char[10]);" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 6, 4, 24, PARAM_TYPE);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void paramTypeClass() {
		init("program Test" + LF +
		     "  class C1 { }" + LF +
		     "  class C2 { }" + LF +
		     "{" + LF +
		     "  void method(C1 c1) { }" + LF +
		     "  void main() {" + LF +
		     "    method(new C2);" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 6, 7, 18, PARAM_TYPE);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void moreParams() {
		init("program Test {" + LF +
		     "  void method(int x, char c) { }" + LF +
		     "  void main() {" + LF +
		     "    method(1, 'a', 1);" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 6, 4, 21, MORE_PARAMS);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void lessParams() {
		init("program Test {" + LF +
		     "  void method(int x, char c) { }" + LF +
		     "  void main() {" + LF +
		     "    method(1);" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 6, 4, 13, LESS_PARAMS);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void incompTypesCond() {
		init("program Test {" + LF +
		     "  void main() int i; {  " + LF +
		     "    if (i > null) { }" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 6, 3, 17, INCOMP_TYPES);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void incompTypesCondArr() {
		init("program Test {" + LF +
		     "  void main() int[] ia; char[] ca; {  " + LF +
		     "    if (ia > ca) { }" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 6, 3, 16, INCOMP_TYPES);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void incompTypesCondClass() {
		init("program Test" + LF +
		     "  class C1 { }" + LF +
		     "{" + LF +
		     "  void main() C1 c1; int i; {  " + LF +
		     "    if (c1 > i) { };" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 6, 5, 15, INCOMP_TYPES);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongEqCheck() {
		init("program Test {" + LF +
		     "  void main() int[] ia1, ia2; {" + LF +
		     "    if (ia1 > ia2) { }" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 6, 3, 18, EQ_CHECK);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noIntOpNeg() {
		init("program Test {" + LF +
		     "  void main() int i; char c; {" + LF +
		     "    i = -c;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 11, NO_INT_OP);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noIntOpPlus() {
		init("program Test {" + LF +
		     "  void main() int i; int[] ia; {" + LF +
		     "    i = i + ia;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 15, NO_INT_OP);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noIntOpTimes() {
		init("program Test {" + LF +
		     "  void main() int i; int[] ia; {" + LF +
		     "    i = ia * i;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 15, NO_INT_OP);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void procAsFunc() {
		init("program Test {" + LF +
		     "  void main() int x; {" + LF +
		     "    x = main();" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 13, INVALID_CALL);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noTypeNew() {
		init("program Test {" + LF +
		     "  void main() int i; {" + LF +
		     "    i = new main;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 17, NO_TYPE);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongArraySize() {
		init("program Test {" + LF +
		     "  void main() int[] ia; {" + LF +
		     "    ia = new int[ia];" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 20, ARRAY_SIZE);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noClassType() {
		init("program Test {" + LF +
		     "  void main() int i; {" + LF +
		     "    i = new int;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 16, NO_CLASS_TYPE);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongFactor() {
		init("program Test {" + LF +
		     "  void main () int i; {  " + LF +
		     "    i = i + if;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 3, 3, 13, INVALID_FACT);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noClass() {
		init("program Test {" + LF +
		     "  void main() int i; {" + LF +
		     "    i = i.i;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 12, NO_CLASS);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noArrayIndex() {
		init("program Test {" + LF +
		     "  void main() int[] ia; {" + LF +
		     "    ia[ia] = 1;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 10, ARRAY_INDEX);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noArray() {
		init("program Test {" + LF +
		     "  void main() int i; {" + LF +
		     "    i[i]++;" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 5, 3, 8, NO_ARRAY);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void wrongRelOp() {
		init("program Test {" + LF +
		     "  void main() int i; {" + LF +
		     "    if (i x 5);" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 3, 3, 11, REL_OP);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noEof() {
		init("program Test {" + LF +
		     "  void main() {}" + LF +
		     "}moretext");
		expectError(EX >= 3, 3, 2, TOKEN_EXPECTED, "end of file");
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void eof1() {
		init("program Test {" + LF +
		     "  void main() {");
		expectError(EX >= 3, 2, 15, TOKEN_EXPECTED, "}");
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void eof2() {
		init("program Test {" + LF +
		     "  void main() {" + LF +
		     "    if ()");
		expectError(EX >= 3, 3, 9, INVALID_FACT);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void eof3() {
		init("program Test" + LF +
		     "  class C {" + LF +
		     "    int i");
		expectError(EX >= 3, 3, 9, TOKEN_EXPECTED, ";");
		parseAndVerify();
	}
	
//  ---- multiple errors & recovery
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void multipleErrors() {
		init("program Test " + LF +
		     "  int x" + LF +
		     "{" + LF +
		     "  void main( {" + LF +
		     "    if (1 x 2);" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 3, 3, 1, TOKEN_EXPECTED, ";");
		expectError(EX >= 4, 4, 14, TOKEN_EXPECTED, ")");
		expectError(EX >= 4, 5, 11, REL_OP);
		parseAndVerify();
	}
	
	@Test(timeout=Configuration.DEFAULT_TIMEOUT)	
	public void recoverDecl1() {
		init("program Test" + LF +
		     "  int i1, if" + LF +
		     "  in i2;" + LF +
		     "  final int i3 = 0;" + LF +
		     "{" + LF +
		     "  void main() {  " + LF +
		     "    if (i1 < i3);" + LF +
		     "    i2 = 0;" + LF +
		     "  }" + LF +
		     "}");

		expectError(EX >= 3, 2, 11, TOKEN_EXPECTED, "identifier");
		expectError(EX >= 5, 8, 8, NOT_FOUND, "i2");
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void recoverDecl2() {
		init("program Test" + LF +
		     "  int i1, if" + LF +
		     "  in i2;" + LF +
		     "  int i3;" + LF +
		     "{" + LF +
		     "  void main() {  " + LF +
		     "    if (i1 < i3);" + LF +
		     "    i2 = 0;" + LF +
		     "  }" + LF +
		     "}");

		expectError(EX >= 3, 2, 11, TOKEN_EXPECTED, "identifier");
		expectError(EX >= 5, 8, 8, NOT_FOUND, "i2");
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void recoverStat() {
		init("program Test {" + LF +
		     "  void main() {  " + LF +
		     "    567 no start on ident so no error reported;" + LF +
		     "    if (1 < 2);" + LF +
		     "    if (1 x 2);" + LF +
		     "  }" + LF +
		     "}");

		expectError(EX == 3, 3, 5, TOKEN_EXPECTED, "}");
		expectError(EX >= 4, 3, 5, INVALID_STAT);
		expectError(EX >= 4, 5, 11, REL_OP);
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noRecover1() {
		init("program Test {" + LF +
		     "  void main this method will never recover");

		expectError(EX >= 3, 2, 13, TOKEN_EXPECTED, "(");
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void noRecover2() {
		init("program Test {" + LF +
		     "  void main() {  " + LF +
		     "    if this method will never recover");

		expectError(EX >= 3, 3, 8, TOKEN_EXPECTED, "(");
		parseAndVerify();
	}

	@Test(timeout=Configuration.DEFAULT_TIMEOUT)
	public void resetErrDist() {
		init("program Test {" + LF +
		     "  void main() {" + LF +
		     "    if () if () if();" + LF +
		     "  }" + LF +
		     "}");
		expectError(EX >= 3, 3, 9, INVALID_FACT);
		expectError(EX >= 4, 3, 15, INVALID_FACT);
		expectError(EX >= 4, 3, 20, INVALID_FACT);
		parseAndVerify();
	}
}
